#include <GL/glut.h>
#include <time.h>
#include <math.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include "fire.h"
#include "voxelsIO.h"

void init()
{
	int i=0;

	globalVolume = loadVolume("blank64.vxl", 1);

	posX=0; posY=0; posZ=0;
	oldAngX=0; oldAngY=0;
	angX=0; angY=0;

	srand(42);

	glEnable(GL_BLEND);
	glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);

	glClearColor(0.7,0.7,0.7,1);
	glShadeModel(GL_SMOOTH);

	glEnable(GL_DEPTH_TEST);

	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();

	gluPerspective(30, 1, 1, 1000);
//	glOrtho(-20.0,20.0,-20.0,20.0,-50.0,50.0);

	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();

	glPointSize(10.0);

	for(i=0; i<MAX_EMITTERS; i++)
	{
		emitters[i].density	 = 255;	// number of new particles created per cycle
		emitters[i].heat	 = 32;	// heat of new particles
		emitters[i].vector_x = 0;	// x co-ordinate of emitter
		emitters[i].vector_y = 0;	// y co-ordinate of emitter
		emitters[i].vector_z = 0;	// z co-ordinate of emitter

		emitters[i].fade=0;			// rate at which rebirth rate decreases

		emitters[i].x=32+i;			// x co-ordinate of emitter
		emitters[i].y=6;			// y co-ordinate of emitter
		emitters[i].z=32;			// z co-ordinate of emitter

		emitters[i].r=0;			// red value
		emitters[i].g=0;			// green value
		emitters[i].b=0;			// blue value
	}
}

void idle(void)
{
	glutPostRedisplay();
}

void display(void)
{
	int				x, y, z,
					current_voxel;
	unsigned char	alpha;

	for(x=0; x<MAX_EMITTERS; x++)
	{
		//get the position of the emitter voxel in the global array
		current_voxel =   emitters[x].x
						+(emitters[x].y*globalVolume->sizeX)
						+(emitters[x].z*globalVolume->sizeX*globalVolume->sizeY);

		//assign the preset values to the emitter
		globalVolume->data[current_voxel].density  = emitters[x].density;
		globalVolume->data[current_voxel].vector_x = emitters[x].vector_x;
		globalVolume->data[current_voxel].vector_y = emitters[x].vector_y;
		globalVolume->data[current_voxel].vector_z = emitters[x].vector_z;
		globalVolume->data[current_voxel].heat = emitters[x].heat;
	}

	glClear(GL_COLOR_BUFFER_BIT|GL_DEPTH_BUFFER_BIT);

	glPushMatrix();

		glTranslatef (0, 0, -200);
		glRotatef (angX, 1, 0, 0);
		glRotatef (angY, 0, 1, 0);

		movement();

//------------------------------------------------------------------------------------------------------------------
//		CALCULTE DENSITY MOVEMENT

		//for every voxel
		for(z=0; z<globalVolume->sizeZ; z++)
			for(y=0;y<globalVolume->sizeY;y++)
				for(x=0;x<globalVolume->sizeX;x++)
				{
					//get the voxel's position in the global array
					current_voxel = (x+(globalVolume->sizeX*y)+(globalVolume->sizeX*globalVolume->sizeY*z));

					//if there is a densty value to move, calculate its movement
					if (globalVolume->data[current_voxel].density!=0)
						moveDensity(x,y,z);
				}

//------------------------------------------------------------------------------------------------------------------
//		DRAW SCENE

		//use GL_POINTS because they're hella fast
		glBegin(GL_POINTS);
			//for every voxel
			for(z=0;z<globalVolume->sizeZ;z++)
				for(y=0;y<globalVolume->sizeY;y++)
					for(x=0;x<globalVolume->sizeX;x++)
					{
						//get the voxel's position in the global array
						current_voxel = (x+(globalVolume->sizeX*y)+(globalVolume->sizeX*globalVolume->sizeY*z));

						//move the values from the buffers to the main memory
						globalVolume->data[current_voxel].heat				= globalVolume->data[current_voxel].heat_buffer;
						globalVolume->data[current_voxel].density			= globalVolume->data[current_voxel].density_buffer;
						globalVolume->data[current_voxel].vector_x			= globalVolume->data[current_voxel].vector_x_buffer;
						globalVolume->data[current_voxel].vector_y			= globalVolume->data[current_voxel].vector_y_buffer;
						globalVolume->data[current_voxel].vector_z			= globalVolume->data[current_voxel].vector_z_buffer;

						//clear the buffers
						globalVolume->data[current_voxel].heat_buffer		= 0;
						globalVolume->data[current_voxel].density_buffer	= 0;
						globalVolume->data[current_voxel].vector_x_buffer	= 0;
						globalVolume->data[current_voxel].vector_y_buffer	= 0;
						globalVolume->data[current_voxel].vector_z_buffer	= 0;

						//calculate alpha using base value plus density plus heat
						alpha = globalVolume->data[current_voxel].alpha
							  + globalVolume->data[current_voxel].density
							  + globalVolume->data[current_voxel].heat;

						//if this is not 0 display the current voxel based on its values and the display mode
						if (alpha != 0)
						{
							if(globalVolume->data[current_voxel].heat > 4)
							{
								if(display_heat==1)
									//sets the colour to yellow if very hot, red if warm and black if cool.
									glColor4ub(255, globalVolume->data[current_voxel].heat * 8, 0, alpha * heat_alpha);
								else if(display_density==1)
									//sets the colour to the colour specified by the emitter.
									glColor4ub(	emitters[0].r,
												emitters[0].g,
												emitters[0].b,
												alpha * density_alpha);
							}
							else if(globalVolume->data[current_voxel].density > 0)
							{
								if(display_density==1)
									//sets the colour to the colour specified by the emitter.
									glColor4ub(	emitters[0].r,
												emitters[0].g,
												emitters[0].b,
												alpha * density_alpha);
							}
							//sets the colour to the base voxel colour.
							else glColor4ub(	globalVolume->data[current_voxel].red,
												globalVolume->data[current_voxel].green,
												globalVolume->data[current_voxel].blue,
												globalVolume->data[current_voxel].alpha * 0.5);

							//draw voxel at appropriate co-ordinate base on how far we are through for loop
							glVertex3f(	x-globalVolume->sizeX/2,
										y-globalVolume->sizeY/2,
										z-globalVolume->sizeZ/2);
						}
					}
		glEnd();

//-------------------------------------------------------------------------------------------------------------------
//		DRAW VECTORS

		if(display_vectors==1)
		{
			//for every voxel
			for(z=0;z<globalVolume->sizeZ;z++)
				for(y=0;y<globalVolume->sizeY;y++)
					for(x=0;x<globalVolume->sizeX;x++)
					{
						//get the voxel's position in the global array
						current_voxel = (x+(globalVolume->sizeX*y)+(globalVolume->sizeX*globalVolume->sizeY*z));

						if (globalVolume->data[current_voxel].density != 0)
						{
						//draw the velocity vector using a GL_LINE from the voxel to an offset proportional to the vector magnitude
						glBegin(GL_LINES);
						glVertex3f( x-globalVolume->sizeX/2,
									y-globalVolume->sizeY/2,
									z-globalVolume->sizeZ/2);
						glVertex3f(	x-globalVolume->sizeX/2+globalVolume->data[current_voxel].vector_x,
									y-globalVolume->sizeY/2+globalVolume->data[current_voxel].vector_y,
									z-globalVolume->sizeZ/2+globalVolume->data[current_voxel].vector_z);

						glEnd();
						}
					}
		}
	glPopMatrix();

	glutSwapBuffers();
}

void moveDensity(short x, short y, short z)
{
	float	theta, phi, u, v,		//for calculating displacement to a point on a sphere
			distance,
			movement,
			dest_vector_x, dest_vector_y, dest_vector_z, dest_density, dest_heat;

	int		source_voxel, dest_voxel,
			samples,
			dest_x, dest_y, dest_z,
			i;

	//get the source voxel's position in the global array
	source_voxel = (x +(globalVolume->sizeX*y) +(globalVolume->sizeX*globalVolume->sizeY*z));
	//work out how many samples are going to be made to shiaft all the density
	samples = (int)globalVolume->data[source_voxel].density / sample_density;

	for(i=0; i<samples; i++)
	{
		//set a random value for distance to determine this samples movement from source voxel
		distance = spread_factor * rand() / (RAND_MAX+1.0);

		u = rand() / (RAND_MAX+1.0);
		v = rand() / (RAND_MAX+1.0);

		//these randomly generated angles will produce evenly spaced points on the surface of a sphere
		theta = 2*PI*u;
		phi = acos(2*v-1);

		movement = (float)globalVolume->data[source_voxel].density * distance;

		//the components of the vector are calculated based.
		//velocity is added as an arbitrary offset
		//density offset is calculated as a trigonometric function of the overall density.
		//The angles generated are used to determine the offset in x, y and z to reach a point on the
		//surface of a sphere of a radius proportional to a random fraction ofthe original density
		dest_vector_x = globalVolume->data[source_voxel].vector_x
					  + sqrt(movement) * sin(theta) * sin(phi);

		dest_vector_y = globalVolume->data[source_voxel].vector_y
					  + globalVolume->data[source_voxel].heat * 0.1
					  + sqrt(movement) * cos(theta) * sin(phi)
					  - mass;

		dest_vector_z = globalVolume->data[source_voxel].vector_z
					  + sqrt(movement) * cos(phi);

		//set the destination to be the source value plus the newly generated vector
		dest_x = x + (int)dest_vector_x;
		dest_y = y + (int)dest_vector_y;
		dest_z = z + (int)dest_vector_z;

		//if the destination is inside the volume
		if(dest_x > 0 && dest_x < globalVolume->sizeX
		&& dest_y > 0 && dest_y < globalVolume->sizeY
		&& dest_z > 0 && dest_z < globalVolume->sizeZ)
		{
			//get the destination voxel's position in the global array
			dest_voxel		= (dest_x +(globalVolume->sizeX*dest_y) +(globalVolume->sizeX*globalVolume->sizeY*dest_z));
			//add the sample density to the destination buffer
			dest_density	= (int)globalVolume->data[dest_voxel].density_buffer + sample_density;
			//alter the destination heat buffer to be the average of its current value plus a fraction of the source heat value
			dest_heat		= ((int)globalVolume->data[source_voxel].heat * cool_factor
							+ (int)globalVolume->data[dest_voxel].heat_buffer)/2;

			//if the density has not moved from its original position
			if (dest_voxel == source_voxel)
			{
				//multiply the values by a variable. this can be used to fade static values
				//or create pyrotechnical effects
				dest_density *= fuel;
				dest_heat *= fuel;
			}

			//if the denstination is inside a solid material move the values back to their original position and delete the velocity value
			if (globalVolume->data[dest_voxel].material == 1)
			{
				dest_voxel 		= source_voxel;
				dest_density	= (int)globalVolume->data[dest_voxel].density_buffer + sample_density;
				dest_heat		= ((int)globalVolume->data[source_voxel].heat * cool_factor
								+ (int)globalVolume->data[dest_voxel].heat_buffer)/2;
				dest_vector_x	= 0;
				dest_vector_y	= 0;
				dest_vector_z	= 0;
			}

			//send the calculated values to the destination buffer.
			globalVolume->data[dest_voxel].heat_buffer     = dest_heat;
			globalVolume->data[dest_voxel].density_buffer  = dest_density;
			globalVolume->data[dest_voxel].vector_x_buffer = dest_vector_x * resistance_factor;
			globalVolume->data[dest_voxel].vector_y_buffer = dest_vector_y * resistance_factor;
			globalVolume->data[dest_voxel].vector_z_buffer = dest_vector_z * resistance_factor;
		}
	}
}

void keyboard(unsigned char key,int x,int y)
{
	keyHit[key]=1; //change value of the array member corresponding to ascii value of key hit
}

void upkeyboard(unsigned char key,int x,int y)
{
	if (key==8)		//if key hit was 'backspace' cycle current control variable
	{
		current_variable++;
		if(current_variable>8)
			current_variable=0;
		switch(current_variable)
		{
			case 0:
				printf("Emitter density   : %d\n",emitters[0].density);
				break;
			case 1:
				printf("Emitter heat      : %d\n",emitters[0].heat);
				break;
			case 2:
				printf("Resistance factor : %f\n",resistance_factor);
				break;
			case 3:
				printf("Cool factor       : %f\n",cool_factor);
				break;
			case 4:
				printf("Spread factor     : %f\n",spread_factor);
				break;
			case 5:
				printf("Mass              : %f\n",mass);
				break;
			case 6:
				printf("Fuel              : %f\n",fuel);
				break;
			case 7:
				printf("Heat alpha        : %f\n",heat_alpha);
				break;
			case 8:
				printf("Density alpha     : %f\n",density_alpha);
				break;
			default:
				break;
		}
	}

	if (keyHit[48])		//if '0' was pressed set sample density to 10
	{
		sample_density = 10;
		printf("Sample density    : %d\n",sample_density);
	}

	if (keyHit[49])		//if '1' was pressed set sample density to 1
	{
		sample_density = 1;
		printf("Sample density    : %d\n",sample_density);
	}

	if (keyHit[50])		//if '2' was pressed set sample density to 2
	{
		sample_density = 2;
		printf("Sample density    : %d\n",sample_density);
	}

	if (keyHit[51])		//if '3' was pressed set sample density to 3
	{
		sample_density = 3;
		printf("Sample density    : %d\n",sample_density);
	}

	if (keyHit[52])		//if '4' was pressed set sample density to 4
	{
		sample_density = 4;
		printf("Sample density    : %d\n",sample_density);
	}

	if (keyHit[53])		//if '5' was pressed set sample density to 5
	{
		sample_density = 5;
		printf("Sample density    : %d\n",sample_density);
	}

	if (keyHit[54])		//if '6' was pressed set sample density to 6
	{
		sample_density = 6;
		printf("Sample density    : %d\n",sample_density);
	}

	if (keyHit[55])		//if '7' was pressed set sample density to 7
	{
		sample_density = 7;
		printf("Sample density    : %d\n",sample_density);
	}

	if (keyHit[56])		//if '8' was pressed set sample density to 8
	{
		sample_density = 8;
		printf("Sample density    : %d\n",sample_density);
	}

	if (keyHit[57])		//if '9' was pressed set sample density to 9
	{
		sample_density = 9;
		printf("Sample density    : %d\n",sample_density);
	}

	if (key==104)		//if key hit was 'h' toggle heat display
	{
		if(display_heat==1)
		{
			display_heat=0;
			printf("Heat              : Off\n");
		}
		else
		{
			display_heat=1;
			printf("Heat              : On\n");
		}
	}

	if (key==112)		//if key hit was 'p' toggle pressure display
	{
		if(display_density==1)
		{
			display_density=0;
			printf("Density           : Off\n");
		}
		else
		{
			display_density=1;
			printf("Density           : On\n");
		}
	}

	if (key==118)		//if key hit was 'v' toggle vertex display
	{
		if(display_vectors==1)
		{
			display_vectors=0;
			printf("Vectors           : Off\n");
		}
		else
		{
			display_vectors=1;
			printf("Vectors           : On\n");
		}
	}

	else
		keyHit[key]=0;	//change value of the array member corresponding to ascii value of key hit
}

void movement()
{
	if (keyHit[27])
	{
//	  free(scene);
		exit(EXIT_SUCCESS);
	}

	if (keyHit[45])	//if '-' was pressed reduce value of current variable
	{
		switch(current_variable)
		{
			case 0:
				if(emitters[0].density > 0) emitters[0].density -= 5;
				printf("Emitter density   : %d\n",emitters[0].density);
				break;
			case 1:
				if(emitters[0].heat > 0) emitters[0].heat -= 1;
				printf("Emitter heat      : %d\n",emitters[0].heat);
				break;
			case 2:
				resistance_factor -= 0.01;
				printf("Resistance factor : %f\n",resistance_factor);
				break;
			case 3:
				cool_factor -= 0.01;
				printf("Cool factor       : %f\n",cool_factor);
				break;
			case 4:
				spread_factor -= 0.01;
				printf("Spread factor     : %f\n",spread_factor);
				break;
			case 5:
				mass -= 0.01;
				printf("Mass              : %f\n",mass);
				break;
			case 6:
				fuel -= 0.01;
				printf("Fuel              : %f\n",fuel);
				break;
			case 7:
				heat_alpha -= 0.01;
				printf("Heat alpha        : %f\n",heat_alpha);
				break;
			case 8:
				density_alpha -= 0.01;
				printf("Density  alpha    : %f\n",density_alpha);
				break;
			default:
				break;
		}
	}

	if (keyHit[61])	//if '=' was pressed increase value of current variable
	{
		switch(current_variable)
		{
			case 0:
				if(emitters[0].density < 255) emitters[0].density += 5;
				printf("Emitter density   : %d\n",emitters[0].density);
				break;
			case 1:
				if(emitters[0].heat < 255) emitters[0].heat += 1;
				printf("Emitter heat      : %d\n",emitters[0].heat);
				break;
			case 2:
				resistance_factor += 0.01;
				printf("Resistance factor : %f\n",resistance_factor);
				break;
			case 3:
				cool_factor += 0.01;
				printf("Cool factor       : %f\n",cool_factor);
				break;
			case 4:
				spread_factor += 0.01;
				printf("Spread factor     : %f\n",spread_factor);
				break;
			case 5:
				mass += 0.01;
				printf("Mass              : %f\n",mass);
				break;
			case 6:
				fuel += 0.01;
				printf("Fuel              : %f\n",fuel);
				break;
			case 7:
				heat_alpha += 0.01;
				printf("Heat alpha        : %f\n",heat_alpha);
				break;
			case 8:
				density_alpha += 0.01;
				printf("Density  alpha    : %f\n",density_alpha);
				break;
			default:
				break;
		}
	}

	if (keyHit[97])		//if 'a' was pressed move emitter left
	{
		if (emitters[0].x > 1) emitters[0].x -= 1;
			printf("Emitter position  : %d %d %d\n",emitters[0].x, emitters[0].y, emitters[0].z);
	}

	if (keyHit[100])	//if 'd' was pressed move emitter right
	{
		if (emitters[0].x < globalVolume->sizeX - 1) emitters[0].x += 1;
			printf("Emitter position  : %d %d %d\n",emitters[0].x, emitters[0].y, emitters[0].z);
	}

	if (keyHit[101])	//if 'e' was pressed move emitter forward
	{
		if (emitters[0].z > 1) emitters[0].z -= 1;
			printf("Emitter position  : %d %d %d\n",emitters[0].x, emitters[0].y, emitters[0].z);
	}

	if (keyHit[105])	//if 'i' was pressed point emitter forward
	{
		if (emitters[0].vector_z > -128) emitters[0].vector_z -= 1;
			printf("Emitter direction : %d %d %d\n",emitters[0].vector_x, emitters[0].vector_y, emitters[0].vector_z);
	}

	if (keyHit[106])	//if 'j' was pressed point emitter left
	{
		if (emitters[0].vector_x > -128) emitters[0].vector_x -= 1;
			printf("Emitter direction : %d %d %d\n",emitters[0].vector_x, emitters[0].vector_y, emitters[0].vector_z);
	}

	if (keyHit[107])	//if 'k' was pressed point emitter backward
	{
		if (emitters[0].vector_z < 127) emitters[0].vector_z += 1;
			printf("Emitter direction : %d %d %d\n",emitters[0].vector_x, emitters[0].vector_y, emitters[0].vector_z);
	}

	if (keyHit[108])	//if 'l' was pressed point emitter right
	{
		if (emitters[0].vector_x < 127) emitters[0].vector_x += 1;
			printf("Emitter direction : %d %d %d\n",emitters[0].vector_x, emitters[0].vector_y, emitters[0].vector_z);
	}

	if (keyHit[113])	//if 'q' was pressed move emitter backward
	{
		if (emitters[0].z < globalVolume->sizeZ - 1) emitters[0].z += 1;
			printf("Emitter position  : %d %d %d\n",emitters[0].x, emitters[0].y, emitters[0].z);
	}

	if (keyHit[115])	//if 's' was pressed move emitter down
	{
		if (emitters[0].y > 1) emitters[0].y -= 1;
			printf("Emitter position  : %d %d %d\n",emitters[0].x, emitters[0].y, emitters[0].z);
	}

	if (keyHit[119])	//if 'w' was pressed move emitter up
	{
		if (emitters[0].y < globalVolume->sizeY - 1) emitters[0].y += 1;
			printf("Emitter position  : %d %d %d\n",emitters[0].x, emitters[0].y, emitters[0].z);
	}

	glutPostRedisplay();
}

void motion(int x, int y)
{
	angX = ( angX + (y - oldAngY) ) % 360 ;
	angY = ( angY + (x - oldAngX) ) % 360 ;
	if (angX >  90)
		angX =  90;
	if (angX < -90)
		angX = -90;
	oldAngX = x;
	oldAngY = y;
	glutPostRedisplay();
}

int main(int argc, char** argv)
{
	glutInit(&argc, argv);
	glutInitDisplayMode (GLUT_DOUBLE|GLUT_RGBA|GLUT_DEPTH);
	glutInitWindowSize (800, 800);
	glutInitWindowPosition (0, 0);
	glutCreateWindow ("FIRE");
	glutDisplayFunc(display);
	glutIdleFunc(idle);
	glutKeyboardFunc(keyboard);
	glutKeyboardUpFunc(upkeyboard);
	glutPassiveMotionFunc(motion);
	init();
	glutMainLoop();

	return 0;
}
