#include <math.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <assert.h>
#include "voxelsIO.h"

volume *loadVolume(char* filename, unsigned char stride)
{
	FILE	*fPointer;
	int		x=0,y=0,z=0;
	int		current_voxel;
	volume	*loadingVolume = (volume*) calloc(1,sizeof(volume));

	if(stride>4 || stride<1)
		return 0;
	loadingVolume->stride=stride;

	// open file
	fPointer = fopen(filename,"rb");
	if(!fPointer)
		return 0;

	fread(&loadingVolume->sizeX,2,1,fPointer);
	fread(&loadingVolume->sizeY,2,1,fPointer);
	fread(&loadingVolume->sizeZ,2,1,fPointer);
	fseek(fPointer,2,SEEK_CUR);

	loadingVolume->num_voxels = loadingVolume->sizeX * loadingVolume->sizeY * loadingVolume->sizeZ;

	loadingVolume->data  = (voxel*) malloc (loadingVolume->num_voxels * sizeof(voxel)
							/ (stride * stride * stride));

	for(z=0; z<loadingVolume->sizeZ; z+=stride)
	{
		for(y=0; y<loadingVolume->sizeY; y+=stride)
		{
			for(x=0; x<loadingVolume->sizeX; x+=stride)
			{
				current_voxel = (x+(loadingVolume->sizeX*y)+(loadingVolume->sizeX*loadingVolume->sizeY*z));

				fread(&loadingVolume->data[current_voxel].red,		sizeof(unsigned char), 1, fPointer);
				fread(&loadingVolume->data[current_voxel].green,	sizeof(unsigned char), 1, fPointer);
				fread(&loadingVolume->data[current_voxel].blue,		sizeof(unsigned char), 1, fPointer);
				fread(&loadingVolume->data[current_voxel].alpha,	sizeof(unsigned char), 1, fPointer);
				fread(&loadingVolume->data[current_voxel].material,	sizeof(unsigned char), 1, fPointer);

				loadingVolume->data[current_voxel].padding			= 0;
				loadingVolume->data[current_voxel].heat				= 0;
				loadingVolume->data[current_voxel].heat_buffer		= 0;
				loadingVolume->data[current_voxel].density			= 0;
				loadingVolume->data[current_voxel].density_buffer	= 0;
				loadingVolume->data[current_voxel].vector_x			= 0;
				loadingVolume->data[current_voxel].vector_y			= 0;
				loadingVolume->data[current_voxel].vector_z			= 0;
				loadingVolume->data[current_voxel].vector_x_buffer	= 0;
				loadingVolume->data[current_voxel].vector_y_buffer	= 0;
				loadingVolume->data[current_voxel].vector_z_buffer	= 0;

				fseek(fPointer, 11+VOXEL_SIZE*(stride-1), SEEK_CUR);
			}
		}
	}

	// close file
	fclose(fPointer);

	return loadingVolume;
}

int newVolume(	short sizeX, short sizeY, short sizeZ,
				char *filename)
{
	FILE	*fPointer;
	short	size_=0;
	char	blankVoxel[VOXEL_SIZE]={0};
	int		x,y,z;

	fPointer = fopen(filename, "wb");		//create a file for storing voxel data
	if(!fPointer)
		return 0;

	//write a 4-bit header telling us the dimensions of the volume.
	fwrite(&sizeX,2,1,fPointer);
	fwrite(&sizeY,2,1,fPointer);
	fwrite(&sizeZ,2,1,fPointer);
	fwrite(&size_,2,1,fPointer);

	//create a blank field for data to be put in later
	for(z=0; z<sizeZ; z++)
		for(y=0; y<sizeY; y++)
			for(x=0; x<sizeX; x++)
				fwrite(blankVoxel,sizeof(unsigned char),(VOXEL_SIZE),fPointer);

	//close the file now that weve finished with it
	fclose(fPointer);

	return 1;
}

int setVoxelColour(	short x, short y, short z,
					unsigned char red, unsigned char green, unsigned char blue, unsigned char alpha,
					char *filename)
{
	FILE	*fPointer;
	short	sizeX,sizeY;

	fPointer = fopen(filename, "rb+");
	if(!fPointer)
		return 0;

	//read file header to find volume dimensions;
	fread(&sizeX,2,1,fPointer);
	fread(&sizeY,2,1,fPointer);

	//move to spedcified co-ordinate
	fseek(fPointer, 8+(VOXEL_SIZE*(x+(sizeX*y)+(sizeX*sizeY*z))), SEEK_SET);

	//write colour at co-ordinate
	fwrite(&red,sizeof(unsigned char),1,fPointer);
	fwrite(&green,sizeof(unsigned char),1,fPointer);
	fwrite(&blue,sizeof(unsigned char),1,fPointer);
	fwrite(&alpha,sizeof(unsigned char),1,fPointer);

	fclose(fPointer);

	return 1;
}

int	setVoxelMaterial(	short x, short y, short z,
						unsigned char mtl,
						char *filename)
{
	FILE	*fPointer;
	short	sizeX,sizeY;

	fPointer = fopen(filename, "rb+");
	if(!fPointer)
		return 0;

	//read file header to find volume dimensions;
	fread(&sizeX,2,1,fPointer);
	fread(&sizeY,2,1,fPointer);

	//move to spedcified co-ordinate
	fseek(fPointer, 12+(VOXEL_SIZE*(x+(sizeX*y)+(sizeX*sizeY*z))), SEEK_SET);

	//write colour at co-ordinate
	fwrite(&mtl,sizeof(unsigned char),1,fPointer);

	fclose(fPointer);

	return 1;
}

